
import json

from django.conf import settings
from django.db.models import signals
from django.db.models.fields import NOT_PROVIDED
from django.db.transaction import atomic
from django.utils.module_loading import import_string

from django_delta_logger import DOES_NOT_EXIST
from django_delta_logger.models import DeltaEvent, DeltaEventType


JsonEncoder = import_string(settings.JSON_ENCODER)


class Cache:
    """
    Functions for keeping and comparing dirty states on django models
    """

    @classmethod
    def update(cls, instance, state=None):
        instance._delta_logger_cache = state or cls.get_current(instance)

    @classmethod
    def compare_and_update(cls, instance):
        cached = getattr(instance, '_delta_logger_cache', cls.get_defaults(instance._meta.model))
        current = cls.get_current(instance)
        delta = cls.get_delta(cached, current)
        cls.update(instance, state=current)
        return delta

    @staticmethod
    def get_delta(a, b):
        return {
            key: new
            for key, old, new
            in (
                (
                    key,
                    a.get(key, DOES_NOT_EXIST),
                    b.get(key, DOES_NOT_EXIST)
                )
                for key
                in set(a.keys()).union(b.keys())
            )
            if (
                old == DOES_NOT_EXIST
                or new == DOES_NOT_EXIST
                or old != new
            )
        }

    @staticmethod
    def get_current(instance):
        return {
            f.attname: getattr(instance, f.attname)
            for f
            in instance._meta.get_fields()
            if hasattr(f, 'attname')
            and getattr(f, 'concrete', False)
            and not getattr(f, 'many_to_many', False)
        }
    
    @staticmethod
    def get_defaults(model):
        return {
            f.attname: None if f.default is NOT_PROVIDED else f.default
            for f
            in model._meta.get_fields()
            if hasattr(f, 'attname')
            and getattr(f, 'concrete', False)
            and not getattr(f, 'many_to_many', False)
        }


class DeltaLogger:
    """
    ContextManager that records changes to the specified django models while active
    """

    def __init__(self, *models, concurrent=False):
        self._models = models
        self._concurrent = concurrent

    def __enter__(self):
        signals.post_init.connect(self.handle_post_init, dispatch_uid='django_delta_logger.post_init')
        signals.pre_save.connect(self.handle_pre_save, dispatch_uid='django_delta_logger.pre_save')
        signals.post_save.connect(self.handle_post_save, dispatch_uid='django_delta_logger.post_save')
        signals.pre_delete.connect(self.handle_pre_delete, dispatch_uid='django_delta_logger.pre_delete')

    def __exit__(self, exc_type, exc_value, traceback):
        signals.post_init.disconnect(self.handle_post_init, dispatch_uid='django_delta_logger.post_init')
        signals.pre_save.disconnect(self.handle_pre_save, dispatch_uid='django_delta_logger.pre_save')
        signals.post_save.disconnect(self.handle_post_save, dispatch_uid='django_delta_logger.post_save')
        signals.pre_delete.disconnect(self.handle_pre_delete, dispatch_uid='django_delta_logger.pre_delete')

    def handle_post_init(self, sender, **kwargs):
        if sender in self._models:
            instance = kwargs.get('instance')
            Cache.update(instance)

    def handle_pre_save(self, sender, **kwargs):
        if (
            sender in self._models
            and self._concurrent
        ):
            instance = kwargs.get('instance')
            existing = instance._meta.model.objects.get(pk=instance.pk)
            Cache.update(instance, state=Cache.get_current(existing))
    
    def handle_post_save(self, sender, **kwargs):
        if sender in self._models:
            instance = kwargs.get('instance')
            created = kwargs.get('created')
            if created:
                Cache.update(instance, state=Cache.get_defaults(instance._meta.model))
            delta = Cache.compare_and_update(instance)
            event = DeltaEvent(
                event_sender=instance,
                event_type=DeltaEventType.CREATED if created else DeltaEventType.UPDATED,
                event_data=json.dumps(delta, cls=JsonEncoder)
            )
            event.save()
    
    def handle_pre_delete(self, sender, **kwargs):
        if sender in self._models:
            instance = kwargs.get('instance')
            event = DeltaEvent(
                event_sender=instance,
                event_type=DeltaEventType.DELETED
            )
            event.save()



import logging

from django.db.models import fields


class IntEnumField(fields.IntegerField):

    description = 'Represents an IntEnum as a model field'

    def __init__(self, enum, *args, **kwargs):
        for item in enum:
            assert isinstance(item.value, int), "IntEnumField only supports integer values in Enums"
        self.enum = enum
        super().__init__(
            *args,
            **{
                'choices': self.get_choices_from_enum(self.enum),
                **kwargs
            }
        )

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        return name, path, [self.enum] + args, kwargs

    def _convert_value_to_enum(self, value):
        if value is not None:
            value_map = self.get_enums_by_value_map(self.enum)
            if value in value_map:
                return value_map[value]
            else:
                logging.WARN('unexpected enum value: {} not found in enum {}'.format(value, self.enum))
                return value
        else:
            return None

    def from_db_value(self, value, expression, connection):
        return self._convert_value_to_enum(value)

    def to_python(self, value):
        if isinstance(value, self.enum):
            return value
        else:
            return self._convert_value_to_enum(value)

    def get_prep_value(self, value):
        if value is None:
            return None
        elif isinstance(value, int):
            return value
        elif hasattr(value, 'value'):
            return value.value
        else:
            raise TypeError('Unable to coerce {} to integer'.format(value))

    def get_db_prep_value(self, value, connection, prepared=False):
        return self.get_prep_value(value)

    @staticmethod
    def get_choices_from_enum(enum):
        return tuple(
            (item.name, item.value)
            for item in enum
        )

    @staticmethod
    def get_enums_by_value_map(enum):
        return {
            item.value: item
            for item in enum
        }



import json

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models

from django_delta_logger import DeltaEventType
from django_delta_logger.fields import IntEnumField


class DeltaEventManager(models.Manager):

    def get_events_for_instance(self, instance, **kwargs):
        content_type = ContentType.objects.get_for_model(instance._meta.model)
        return self.filter(
            event_sender_type=content_type,
            event_sender_id=instance.pk,
            **kwargs
        )
    
    def get_events_for_model(self, model, **kwargs):
        content_type = ContentType.objects.get_for_model(model)
        return self.filter(event_sender_type=content_type, **kwargs)


class DeltaEvent(models.Model):
    objects = DeltaEventManager()

    event_sender_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    event_sender_id = models.PositiveIntegerField()
    event_sender = GenericForeignKey('event_sender_type', 'event_sender_id')
    event_time = models.DateTimeField(auto_now_add=True)
    event_type = IntEnumField(DeltaEventType)
    event_data = models.TextField(blank=True, null=True)

    def __data_view__(self):
        """
        Used when serializing
        """
        if self.event_data:
            data = json.loads(self.event_data)
            keys = list(data.keys())
        else:
            data, keys = None, None
        return {
            'operation': self.event_type.name.lower(),
            'changed': keys,
            'data': data,
            'pk': self.pk,
            'class': self.event_sender_type.model_class()._meta.label,
            'time': self.event_time,
        }




__VERSION__ = '0.1.0'


from enum import Enum
import json

from django.core.exceptions import ImproperlyConfigured


default_app_config = 'django_delta_logger.apps.DjangoDeltaLoggerConfig'


DOES_NOT_EXIST = object()
DeltaEventType = Enum('DeltaEventType', 'CREATED UPDATED DELETED')



from enum import Enum

from django.core.serializers.json import DjangoJSONEncoder


class JsonEncoder(DjangoJSONEncoder):
    def default(self, o):
        if hasattr(o, '__data_view__') and callable(o.__data_view__):
            return o.__data_view__()
        if isinstance(o, Enum):
            return o.value
        else:
            return super().default(o)



import setuptools


with open('README.md', 'r') as file_handle:
    long_description = file_handle.read()


setuptools.setup(
    name='django_delta_logger',
    version='0.1.0',
    author='Tanner Netterville',
    author_email='tannern@gmail.com',
    description='Model instance change logging for django models.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url=None,
    package_dir={'': 'src'},
    packages=setuptools.find_packages(where='src'),
    install_requires=(
        'django>=2.2',
    ),
    test_suite='tests.run_test_suite',
    classifiers=(
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    )
)



# Generated by Django 2.2.3 on 2019-07-31 01:53

from django.db import migrations, models
import django.db.models.deletion
import django_delta_logger
import django_delta_logger.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='DeltaEvent',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('event_sender_id', models.PositiveIntegerField()),
                ('event_time', models.DateTimeField(auto_now_add=True)),
                ('event_type', django_delta_logger.fields.IntEnumField(django_delta_logger.DeltaEventType, choices=[('CREATED', 1), ('UPDATED', 2), ('DELETED', 3)])),
                ('event_data', models.TextField(blank=True, null=True)),
                ('event_sender_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
            ],
        ),
    ]
